module Comparable exposing
    ( Comparable, default
    , lt, gt, lte, gte, max, min
    , maximum, minimum, sort
    )

{-| This module contains replacements for functions that are imported by default.


# Comparable

@docs Comparable, default


# Basics

Basic functions with explicit comparison.

@docs lt, gt, lte, gte, max, min


# List

List functions with explicit comparison.

@docs maximum, minimum, sort

-}

import Basics
import List


{-| A function to implement comparison for a given type of value.
-}
type alias Comparable value =
    value -> value -> Basics.Order


{-| The default implementation for comparison of values in Elm.
This function implements equality as well as ordering for some types of values - Custom types aren't supported.
-}
default : Comparable comparable
default =
    Basics.compare



-- Basics


{-| is the first value less than the second?
-}
lt : Comparable value -> value -> value -> Bool
lt compare first second =
    compare first second == Basics.LT


{-| is the first value greater than the second?
-}
gt : Comparable value -> value -> value -> Bool
gt compare first second =
    compare first second == Basics.GT


{-| is the first value less than or equal to the second?
-}
lte : Comparable value -> value -> value -> Bool
lte compare first second =
    compare first second /= Basics.GT


{-| is the first value greater than or equal to the second?
-}
gte : Comparable value -> value -> value -> Bool
gte compare first second =
    compare first second /= Basics.LT


{-| return the greater of two values.
-}
max : Comparable value -> value -> value -> value
max compare first second =
    case compare first second of
        Basics.GT ->
            first

        _ ->
            second


{-| return the lesser of two values.
-}
min : Comparable value -> value -> value -> value
min compare first second =
    case compare first second of
        Basics.LT ->
            first

        _ ->
            second



-- List


{-| Find the maximum element in a non-empty list.
-}
maximum : Comparable value -> List value -> Maybe value
maximum compare list =
    case list of
        x :: xs ->
            Just (List.foldl (max compare) x xs)

        _ ->
            Nothing


{-| Find the minimum element in a non-empty list.
-}
minimum : Comparable value -> List value -> Maybe value
minimum compare list =
    case list of
        x :: xs ->
            Just (List.foldl (min compare) x xs)

        _ ->
            Nothing


{-| Sort values Comparable lowest to highest.
-}
sort : Comparable value -> List value -> List value
sort compare xs =
    List.sortWith compare xs
